#version 150 core

in vec2 textureCoord;

uniform sampler2D albedoFirstBuffer;
uniform sampler2D normalFirstBuffer;
uniform sampler2D lightFirstBuffer;
uniform sampler2D ssaoBuffer;
uniform sampler2D depthFirstBuffer;
uniform sampler2D bentNormalBuffer;
uniform sampler2D heightBuffer;
uniform sampler2D depthSecondBuffer;

uniform int   drawMode;
uniform float gammaCorrection;
uniform float exposure;
uniform bool  drawLights;
uniform bool  drawSSAO;
uniform vec3 ambientIntensity;

uniform bool homogeneousReversedDepth;

uniform int mipmapLevel;

out vec4 color;

vec3 screenGamma = vec3( 1.0f / gammaCorrection );

float linearizeDepth( in float z ) {
  if ( homogeneousReversedDepth ) {
    return z;
  }
  float n = 1.0;    // camera z near
  float f = 100.0;  // camera z far
  return ( 2.0 * n ) / ( f + n - z * ( f - n ) );
}

// Returns +-1
vec2 signNotZero( in vec2 v ) {
  return vec2( ( v.x >= 0.0 ) ? 1.0 : -1.0, ( v.y >= 0.0 ) ? 1.0 : -1.0 );
}
// Assume normalized input. Output is on [-1, 1] for each component.
vec2 encode( in vec3 v ) {
  // Project the sphere onto the octahedron, and then onto the xy plane
  vec2 p = v.xy * ( 1.0 / ( abs( v.x ) + abs( v.y ) + abs( v.z ) ) );
  // Reflect the folds of the lower hemisphere over the diagonals
  return ( v.z <= 0.0 ) ? ( ( 1.0 - abs( p.yx ) ) * signNotZero( p ) ) : p;
}
vec3 decode( in vec2 e ) {
  vec3 v = vec3( e.xy, 1.0 - abs( e.x ) - abs( e.y ) );
  if ( v.z < 0 ) {
    v.xy = ( 1.0 - abs( v.yx ) ) * signNotZero( v.xy );
  }
  return normalize( v );
}

void main() {
  ivec2 ssCoord = ivec2( gl_FragCoord.xy );

  if ( drawMode == 1 ) {
    vec4 albedoFirst = texelFetch( albedoFirstBuffer, ssCoord, 0 );
    vec3 colorFirst  = albedoFirst.rgb;
    if ( drawLights ) {
      vec3  lightFirst       = texelFetch( lightFirstBuffer, ssCoord, 0 ).xyz;
      vec3  selfIllumination = vec3( texelFetch( normalFirstBuffer, ssCoord, 0 ).a );
      float ssao             = 1.0f;
      if ( drawSSAO ) {
        ssao = texelFetch( ssaoBuffer, ssCoord, 0 ).x;
      }
      colorFirst *= max( lightFirst + ambientIntensity * ssao, selfIllumination );
    }

    /*if ( drawSecondGeometryLayer ) {
			if ( albedoFirst.a < 1.0f ) {
				vec3 colorSecond = texelFetch( albedoSecondBuffer, ssCoord, 0 ).rgb;
				if ( drawLights ) {
					vec3 lightSecond = texelFetch( lightSecondBuffer, ssCoord, 0 ).xyz;
					colorSecond     *= lightSecond + ambientIntensity;
				}
				colorFirst = mix( colorSecond, colorFirst, albedoFirst.a );
			}
		}*/

    // Exposure tone mapping
    vec3 mapped = vec3( 1.0f ) - exp( -colorFirst * exposure );
    color       = vec4( pow( mapped, screenGamma ), 1.0f );
  } else if ( drawMode == 2 ) {
    //color = vec4( decode( texelFetch( normalFirstBuffer, ssCoord, 0 ).xy ), 1.0f );
    color = vec4( texelFetch( normalFirstBuffer, ssCoord, 0 ).xyz, 1.0f );
  } else if ( drawMode == 3 ) {
    float ssao = 1.0f;
    if ( drawSSAO ) {
      ssao = texelFetch( ssaoBuffer, ssCoord, 0 ).x;
    }
    vec3 light = texelFetch( lightFirstBuffer, ssCoord, 0 ).rgb + ( ambientIntensity * ssao );
    color      = vec4( pow( light, screenGamma ), 1.0f );
  } else if ( drawMode == 4 ) {
    vec3 ssaoValue = vec3( texelFetch( ssaoBuffer, ssCoord, 0 ).x );
    /*if (abs(ssaoValue.x - 0.5) < 0.01) {
			ssaoValue = vec3(0,1,0);
		}
		if (abs(ssaoValue.x - 0.25) < 0.01) {
			ssaoValue = vec3(1,0,0);
		}*/
    color = vec4( ssaoValue, 1.0f );
  } else if ( drawMode == 5 ) {
    color = vec4( vec3( linearizeDepth( textureLod( depthFirstBuffer, textureCoord, mipmapLevel ).r ) ), 1.0f );
    //color = vec4( vec3( textureLod( depthFirstBuffer, textureCoord, mipmapLevel ).r ), 1.0f );
    //color = vec4( vec3( texelFetch( normalFirstBuffer, ssCoord, 0 ).a ), 1.0f );
  } else if ( drawMode == 6 ) {
    color = texelFetch( bentNormalBuffer, ssCoord, 0 );
  } else if ( drawMode == 7 ) {
    color = vec4( vec3( linearizeDepth( textureLod( heightBuffer, textureCoord, mipmapLevel ).g ) ), 1.0f );
  } else if ( drawMode == 8 ) {
    color = vec4( vec3( linearizeDepth( textureLod( depthSecondBuffer, textureCoord, mipmapLevel ).r ) ), 1.0f );
  }
}
