#version 150 core
#extension GL_ARB_explicit_attrib_location : enable

in vec2 textureCoord;
in mat3 tbnMatrix;
in mat4 tangentToClipMatrix;
in vec3 wsPosition;
in vec3 tsPosition;
in vec3  tsCameraPosition;
in float distanceToFragment;

uniform sampler2D textureAlbedo;
uniform sampler2D textureNormal;
uniform sampler2D textureHeight;
uniform sampler2D depthBufferFirst;
uniform sampler2D depthBufferSecond;

// флаг вывода текста
uniform bool isColor;

uniform bool drawPOM;
// цвет текста
uniform vec3 color;

uniform float alphaTestThreshold;

// переданные параметры полигона
uniform float transparency;
uniform float selfIllumination;

uniform bool homogeneousReversedDepth;
uniform mat4 viewProjectionInversedMatrix;

uniform vec3 wsCameraPosition;
uniform vec2 screenSize;

// Ouput data
layout( location = 0 ) out vec4 albedoBuffer;
layout( location = 1 ) out vec4 normalBuffer;
layout( location = 2 ) out vec2 heightBuffer;

// Returns +-1
vec2 signNotZero( in vec2 v ) {
  return vec2( ( v.x >= 0.0 ) ? 1.0 : -1.0, ( v.y >= 0.0 ) ? 1.0 : -1.0 );
}
// Assume normalized input. Output is on [-1, 1] for each component.
vec2 encode( in vec3 v ) {
  // Project the sphere onto the octahedron, and then onto the xy plane
  vec2 p = v.xy * ( 1.0 / ( abs( v.x ) + abs( v.y ) + abs( v.z ) ) );
  // Reflect the folds of the lower hemisphere over the diagonals
  return ( v.z <= 0.0 ) ? ( ( 1.0 - abs( p.yx ) ) * signNotZero( p ) ) : p;
}
vec3 decode( in vec2 e ) {
  vec3 v = vec3( e.xy, 1.0 - abs( e.x ) - abs( e.y ) );
  if ( v.z < 0 ) {
    v.xy = ( 1.0 - abs( v.yx ) ) * signNotZero( v.xy );
  }
  return normalize( v );
}

// получаем однородные координаты экрана из координат окна
vec2 getSSPosition( in vec2 windowSpacePosition ) {
  return windowSpacePosition / screenSize;
}
// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec4 getWSPosition( in vec2 ssPosition, in float z ) {
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    z = z * 2.0f - 1.0f;
  }
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssPosition * 2.0f - 1.0f, z, 1.0f );
  return wsPosition / wsPosition.w;
}

float getHeight( in vec2 texCoord ) {
  return 1.0f - texture( textureHeight, texCoord ).x;
}
// возвращает величину сдвига в tangent space относительно исходных текстурных координат
vec3 parallaxOcclusionMapping( in vec2 texCoord ) {
  float currentHeightMapValue = getHeight( texCoord );
  // POM имеет смысл считать только для фрагментов с ненулевой высотой
  if ( currentHeightMapValue <= 0.03f ) {
    return vec3( 0.0f, 0.0f, 0.0f );
  }

  vec3  cameraDirection = normalize( tsPosition - tsCameraPosition );
  float heightScale     = 0.01f;
  // вычисляем количество слоев, в зависимости от угла наклона камеры относительно поверхности
  const float minLayers = 16;
  const float maxLayers = 32;
  int         numLayers = int( mix( maxLayers, minLayers, abs( dot( vec3( 0.0f, 0.0f, 1.0f ), cameraDirection ) ) ) );
  // высота одного слоя
  float layerHeight = 1.0f / numLayers;
  // величина сдвига текстурных координат для каждого слоя
  vec2 deltaTexCoords = layerHeight * heightScale * ( cameraDirection.xy / cameraDirection.z );

  vec2  currentTexCoords   = texCoord;
  float currentLayerHeight = 0.0f;
  vec2  prevTexCoords      = currentTexCoords;
  float prevHeightMapValue = currentHeightMapValue;
  while ( currentLayerHeight < currentHeightMapValue ) {
    // обновляем значения предыдущего шага
    prevTexCoords      = currentTexCoords;
    prevHeightMapValue = currentHeightMapValue;
    // высота текущего слоя
    currentLayerHeight += layerHeight;
    // сдвигаем текстурные координаты
    currentTexCoords -= deltaTexCoords;
    // новое значение высоты
    currentHeightMapValue = getHeight( currentTexCoords );
  }
  // считаем высоты текущего и предыдущего слоя и делаем линейную интерполяцию текстурных координат
  float currentHeight  = currentHeightMapValue - currentLayerHeight;
  float previousHeight = prevHeightMapValue - currentLayerHeight + layerHeight;
  float weight         = currentHeight / ( currentHeight - previousHeight );
  currentTexCoords     = prevTexCoords * weight + currentTexCoords * ( 1.0f - weight );
  // интерполяция высоты
  float interpolatedHeight = currentLayerHeight + previousHeight * weight + currentHeight * ( 1.0f - weight );

  // TODO: подумать, можно ли тут избавиться от вычитания, сразу посчитав величину сдвига
  return vec3( texCoord - currentTexCoords, interpolatedHeight * 0.1f );
}

vec3 getNormal( in vec2 texCoord ) {
  vec3 tsNormal = texture( textureNormal, texCoord ).xyz;
  // переводим в NDC [-1.0f, 1.0f]
  vec3 ndcNormal = tsNormal * 2.0f - 1.0f;
  // переводим в мировое пространство
  return normalize( tbnMatrix * ndcNormal );
}

void main() {
  const float drawPOMDistance = 45.0f;
  vec3        tsShift         = vec3( 0.0f, 0.0f, 0.0f );
  vec2        texCoord        = textureCoord;
  if ( drawPOM && distanceToFragment < drawPOMDistance ) {
    // получаем смещение текстурных координат
    tsShift = parallaxOcclusionMapping( texCoord );
    texCoord -= tsShift.xy;
  }

  heightBuffer.x = texelFetch( depthBufferFirst, ivec2( gl_FragCoord.xy ), 0 ).x;
  heightBuffer.y = texelFetch( depthBufferSecond, ivec2( gl_FragCoord.xy ), 0 ).x;

  normalBuffer = vec4( getNormal( texCoord ), selfIllumination );

  if ( isColor ) {
    albedoBuffer = vec4( color, 1.0f );
    return;
  }
  if ( drawPOM && distanceToFragment < drawPOMDistance && tsShift.z != 0.0f ) {
    // записываем значение глубины на основе карты высот текстуры
    vec4 tsNewDepthPosition = vec4( tsPosition - tsShift, 1.0f );
    vec4 csNewDepthPosition = tangentToClipMatrix * tsNewDepthPosition;
    heightBuffer.x          = ( csNewDepthPosition.z / csNewDepthPosition.w );
    if ( !homogeneousReversedDepth ) {
      heightBuffer.x = heightBuffer.x * 0.5f + 0.5f;
    }

    /*
		// определяем силуэты для POM
		const float deltaCoefficient = 5.0f;
		const float positionThreshold = 1.7f;
		// z-координату оставляем исходную, поскольку нам нужно искать точку в плоскости текущего полигона
		// tangent space
		vec4 tsNewPosition = vec4( tsPosition.xy - tsShift.xy * deltaCoefficient, tsPosition.z, 1.0f );
		// clip space
		vec4 csNewPosition = tangentToClipMatrix * tsNewPosition;
		// [0,1] screen space
		vec2 ssNewPosition = ( csNewPosition.xy / csNewPosition.w ) * 0.5f + 0.5f;
		// глубина нового пикселя
		float newDepth = textureLod( depthBufferFirst, ssNewPosition, 0 ).x;
		// если значение есть в буфере (не вышли за границу экрана)
		if ( newDepth != 0.0f ) {
			vec3 wsNewPosition = getWSPosition( ssNewPosition, newDepth ).xyz;
			vec3 wsNewDirection = wsNewPosition - wsPosition;
			// отбрасываем пиксель, если расстояние между точками значительное
			// и отбрасываемая часть полигона находится перед остальной геометрией, а не позади
			// (векторы из исходного пикселя и из камеры, направленные в новый пиксель - сонаправлены)
			if ( length( wsNewDirection ) > positionThreshold && dot( wsNewDirection, wsNewPosition - wsCameraPosition ) > 0.0f ) {
				albedoBuffer = vec4( 0.0f, 0.0f, 0.0f, 0.0f );
				return;
			}
		}*/
  }

  vec4  albedoSample = texture( textureAlbedo, texCoord );
  float alpha        = min( albedoSample.a, transparency );
  if ( alpha < alphaTestThreshold ) {
    discard;
  }
  albedoBuffer = vec4( albedoSample.rgb, alpha );
}
