#version 150 core

uniform sampler2D depthBuffer;
uniform sampler2D normalBuffer;
uniform sampler2D bentNormalBuffer;
uniform samplerCubeShadow shadowCubemapBuffer;
uniform sampler2DShadow shadowBuffer;

uniform mat4 lightProjectionMatrix;

// для восстановления позиции из глубины
uniform bool homogeneousReversedDepth;
uniform mat4 viewProjectionMatrix;
uniform mat4 viewProjectionInversedMatrix;

// для преобразования вектора в пространстве мира в глубину в пространстве света
//uniform vec2 lightProjectionInfo;
uniform bool softShadows;
uniform bool contactShadows;

uniform vec2 screenSize;

uniform bool drawLights;
uniform bool drawShadows;
uniform bool drawSSAO;

uniform int lightType;
uniform vec3 wsLightPosition;
//uniform vec3 ssLightPosition;
uniform vec3 lightColor;
// uniform float lightBrightness;
uniform float lightRadius;
uniform float lightAttenuationConstant;
uniform float lightAttenuationLinear;
uniform float lightAttenuationQuadratic;
uniform vec3  wsLightDirection;
uniform float lightConeInner;
uniform float lightConeOuter;

uniform vec3 wsCameraPosition;
uniform vec3 wsCameraDirection;

out vec3 fragmentColor;

//Sphere of samples for the World Space sampling
const vec3 poissonSphere[ 16 ] = vec3[]( vec3( 0.53812504, 0.18565957, -0.43192 ), vec3( 0.13790712, 0.24864247, 0.44301823 ),
  vec3( 0.33715037, 0.56794053, -0.005789503 ), vec3( -0.6999805, -0.04511441, -0.0019965635 ),
  vec3( 0.06896307, -0.15983082, -0.85477847 ), vec3( 0.056099437, 0.006954967, -0.1843352 ), vec3( -0.014653638, 0.14027752, 0.0762037 ),
  vec3( 0.010019933, -0.1924225, -0.034443386 ), vec3( -0.35775623, -0.5301969, -0.43581226 ), vec3( -0.3169221, 0.106360726, 0.015860917 ),
  vec3( 0.010350345, -0.58698344, 0.0046293875 ), vec3( -0.08972908, -0.49408212, 0.3287904 ),
  vec3( 0.7119986, -0.0154690035, -0.09183723 ), vec3( -0.053382345, 0.059675813, -0.5411899 ),
  vec3( 0.035267662, -0.063188605, 0.54602677 ), vec3( -0.47761092, 0.2847911, -0.0271716 ) );

// Returns +-1
vec2 signNotZero( in vec2 v ) {
  return vec2( ( v.x >= 0.0 ) ? 1.0 : -1.0, ( v.y >= 0.0 ) ? 1.0 : -1.0 );
}
// Assume normalized input. Output is on [-1, 1] for each component.
vec2 encode( in vec3 v ) {
  // Project the sphere onto the octahedron, and then onto the xy plane
  vec2 p = v.xy * ( 1.0 / ( abs( v.x ) + abs( v.y ) + abs( v.z ) ) );
  // Reflect the folds of the lower hemisphere over the diagonals
  return ( v.z <= 0.0 ) ? ( ( 1.0 - abs( p.yx ) ) * signNotZero( p ) ) : p;
}
vec3 decode( in vec2 e ) {
  vec3 v = vec3( e.xy, 1.0 - abs( e.x ) - abs( e.y ) );
  if ( v.z < 0 ) {
    v.xy = ( 1.0 - abs( v.yx ) ) * signNotZero( v.xy );
  }
  return normalize( v );
}

vec3 ssToNDC( in vec3 ssPosition ) {
  vec3 ndcPosition = vec3( ssPosition.xy * 2.0f - 1.0f, ssPosition.z );
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    ndcPosition.z = ndcPosition.z * 2.0f - 1.0f;
  }
  return ndcPosition;
}
vec3 ndcToSS( in vec3 ndcPosition ) {
  vec3 ssPosition = vec3( ndcPosition.xy * 0.5f + 0.5f, ndcPosition.z );
  if ( !homogeneousReversedDepth ) {
    // если глубина была в диапазоне [-1, 1]
    ssPosition.z = ssPosition.z * 0.5f + 0.5f;
  }
  return ssPosition;
}
// получаем однородные координаты экрана из координат окна
vec3 getSSPosition( in vec2 windowSpacePosition, in float z ) {
  return vec3( windowSpacePosition / screenSize, z );
}
// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec3 ssToWSPosition( in vec3 ssPosition ) {
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssToNDC( ssPosition ), 1.0f );
  return wsPosition.xyz / wsPosition.w;
}
vec3 getWSPositionFromDepth( in vec2 windowSpacePosition ) {
  float z = texelFetch( depthBuffer, ivec2( windowSpacePosition ), 0 ).x;
  return ssToWSPosition( getSSPosition( windowSpacePosition, z ) );
}

vec3 wsToSSPosition( in vec3 wsPosition ) {
  vec4 csPosition = viewProjectionMatrix * vec4( wsPosition, 1.0f );
  if ( csPosition.w < 0.0f ) {
    csPosition.w = -csPosition.w;
  }
  vec3 ndcPosition = csPosition.xyz;
  if ( csPosition.w != 0.0f ) {
    ndcPosition /= csPosition.w;
  }
  return ndcToSS( ndcPosition );
}

// проверяем, лежит ли точка за пределами видимой области
bool isClipped( in vec3 ndcPosition ) {
  if ( ndcPosition.x < -1.0f || ndcPosition.x > 1.0f || ndcPosition.y < -1.0f || ndcPosition.y > 1.0f ||
    ( homogeneousReversedDepth && ndcPosition.z < 0.0f ) || ( !homogeneousReversedDepth && ndcPosition.z < -1.0f ) ||
    ndcPosition.z > 1.0f ) {
    return true;
  }
  return false;
}

// Let V be the light-to-fragment direction vector.
// V needs to be in world space since the shadow cube map is normally axis aligned in world space
// thus the "max(abs(V.x),abs(V.y),abs(V.z))"-part only works if V is a world space direction vector.
// The Z value in the respective cube side frustum/"eye space" can be calculated by taking the max of the absolute values of V's components.
// Z = max(abs(V.x),abs(V.y),abs(V.z))
// Then, to be precise, we should negate Z because in OpenGL, the negative Z-axis points into the screen/view frustum.
// Now we want to get the depth buffer "compatible" value of that -Z.
// Looking at the OpenGL perspective matrix we see that, for any homogeneous vector multiplied with that matrix,
// the resulting z value is completely independent of the vector's x and y components.
// So we can simply multiply this matrix with the homogeneous vector (0,0,-Z,1) and we get the vector (components):
// x = 0
// y = 0
// z = (-Z * -(f+n) / (f-n)) + (-2*f*n / (f-n))
// w = Z
// Then we need to do the perspective divide, so we divide z by w (Z) which gives us:
// z' = (f+n) / (f-n) - 2*f*n / (Z* (f-n))
// This z' is in OpenGL's normalized device coordinate (NDC) range [-1,1] and needs to be transformed into a depth buffer compatible range of [0,1]:
// z_depth_buffer_compatible = (z' + 1.0) * 0.5
float vectorToDepth( in vec3 fragmentToLight ) {
  vec3  absVector = abs( fragmentToLight );
  float lsDepth   = max( absVector.x, max( absVector.y, absVector.z ) );
  // clip space
  vec4 csPosition = lightProjectionMatrix * vec4( 0.0f, 0.0f, -lsDepth, 1.0f );
  // NDC depth
  float ndcDepth = csPosition.z / csPosition.w;

  if ( !homogeneousReversedDepth ) {
    // если clip space в диапазоне [-1, 1], преобразуем в [0, 1]
    return ndcDepth * 0.5f + 0.5f;
  }
  return ndcDepth;
}

/*vec2 VogelDiskSample( int sampleIndex, int numSamples, float spinAngle ) {
	const float goldenAngle = 2.4f;

	float r = sqrt( ( float(sampleIndex) + 0.5f ) / numSamples );
	float theta = float(sampleIndex) * goldenAngle + spinAngle;

	return r * vec2(cos(theta), sin(theta));
}*/

/*float AvgBlockersDepthToPenumbra( float lightSize, float z_shadowMapView, float avgBlockersDepth ) {
	return lightSize * ( z_shadowMapView - avgBlockersDepth ) / avgBlockersDepth;
}*/
/*float AvgBlockersDepthToPenumbra(float z_shadowMapView, float avgBlockersDepth)
{
  float penumbra = (z_shadowMapView - avgBlockersDepth) / avgBlockersDepth;
  penumbra *= penumbra;
  return clamp(80.0f * penumbra, 0.0, 1.0);
}*/
/*float Penumbra( float gradientNoise, vec2 shadowMapUV, float z_shadowMapView, int samplesCount ) {
	float avgBlockersDepth = 0.0f;
	float blockersCount = 0.0f;

	for ( int i = 0; i < samplesCount; i++ ) {
		vec2 sampleUV = VogelDiskSample( i, samplesCount, gradientNoise );
		sampleUV = shadowMapUV + penumbraFilterMaxSize * sampleUV;

		float sampleDepth = shadowMapTexture.SampleLevel( pointClampSampler, sampleUV, 0 ).x;

		if ( sampleDepth < z_shadowMapView ) {
			avgBlockersDepth += sampleDepth;
			blockersCount += 1.0f;
		}
	}

	if ( blockersCount > 0.0f ) {
		avgBlockersDepth /= blockersCount;
		return AvgBlockersDepthToPenumbra( z_shadowMapView, avgBlockersDepth );
	} else {
		return 0.0f;
	}
}*/

// Квадратичное затухание
float quadraticAttenuation( in float lightDistance, in float lightRadius, in float lightAttenuationConstant ) {
  // расчитываем затухание только в пределах радиуса
  float d         = min( lightRadius, lightDistance );
  float nominator = clamp( 1.0f - pow( d / lightRadius, 4 ), 0.0f, 1.0f );
  // делимое в квадрате для более плавного затухания
  nominator *= nominator;
  // квадратичное затухание с параметрами
  float denominator = lightAttenuationConstant + lightAttenuationLinear * d + lightAttenuationQuadratic * d * d;
  return nominator / denominator;

  /*// ue4 - 77.1
	float d         = min( lightRadius, lightDistance );
	float nominator = clamp( 1.0f - pow( d / lightRadius, 4 ), 0.0f, 1.0f );
	// делимое в квадрате для более плавного затухания + учет яркости источника
	nominator *= nominator;
	// квадратичное затухание с параметрами
	float denominator = d * d + lightAttenuationConstant;
	return nominator / denominator;*/

  // хороший, быстрый и светлый вариант - 75.7
  //return max( 1.0 / sqrt( lightDistance ) - 1.0 / sqrt( lightRadius ), 0.0 );

  /*// светлый вариант 76.7
	float distOverRadius = sqrt( sqrt( lightDistance / lightRadius ) );
	return sqrt( clamp( 1.0 - distOverRadius, 0.0f, 1.0f ) ) / ( sqrt( lightDistance ) + 1.0 );*/
}
float quadraticAttenuationFast( in float lightDistance, in float lightRadius ) {
  return max( 1.0 / sqrt( lightDistance ) - 1.0 / sqrt( lightRadius ), 0.0 );
}

// 79.3
float calculateCubemapShadow( in vec3 wsPosition, in vec3 wsFragToLight, in vec3 wsFragToLightDir, in float lightDistance ) {
  float shadowCoefficient = 1.0f;
  if ( softShadows ) {
    // PCF
    int   samples    = 16;
    float diskRadius = 0.5f;

    // TODO: это всё переделать
    for ( int i = 0; i < samples; ++i ) {
      vec3  newLightDirection = -wsFragToLight + poissonSphere[ i ] * diskRadius;
      float fragmentDepth     = vectorToDepth( newLightDirection );
      shadowCoefficient += texture( shadowCubemapBuffer, vec4( newLightDirection, fragmentDepth ) );
    }
    shadowCoefficient /= float( samples );

    /*float penumbra = Penumbra( gradientNoise, shadowMapUV, z_shadowMapView, 16 );

		float shadow = 0.0f;
		for ( int i = 0; i < 16; i++ ) {
			vec2 sampleUV = VogelDiskOffset( i, 16, gradientNoise );
			sampleUV = shadowMapUV + sampleUV * penumbra * shadowFilterMaxSize;

			shadow += shadowMapTexture.SampleCmp( linearClampComparisonSampler, sampleUV, z_shadowMapView ).x;
		}
		shadow /= 16.0f;*/
  } else {
    // simple shadow
    float fragmentDepth = vectorToDepth( wsFragToLight );
    shadowCoefficient   = texture( shadowCubemapBuffer, vec4( -wsFragToLight, fragmentDepth ) );
  }

  if ( contactShadows && shadowCoefficient != 0.0f ) {
    // screen space contact shadows
    float radius            = 0.3f;
    float numSteps          = 32.0f;
    float sampleStep        = radius / numSteps;
    vec3  positionStep      = wsFragToLightDir * sampleStep;
    vec3  wsCurrentPosition = wsPosition;

    float shadowMultiplier       = 0.0f;
    int   numSamplesUnderSurface = 0;

    // расстояние от камеры до фрагмента сцены
    //float distanceToFragment = length( wsPosition - wsCameraPosition );

    for ( int i = 0; i < numSteps; ++i ) {
      wsCurrentPosition += positionStep * i;
      //if ( length( wsCurrentPosition - wsPosition ) >= length( wsCurrentPosition - wsLightPosition ) ) {
      //	break;
      //}
      vec4 csPosition = viewProjectionMatrix * vec4( wsCurrentPosition, 1.0f );
      // если вышли за область видимости, прерываем обход
      if ( csPosition.w < 0.0f ) {
        break;
      }
      vec3 ndcPosition = csPosition.xyz;
      if ( csPosition.w != 0.0f ) {
        ndcPosition /= csPosition.w;
      }
      if ( isClipped( ndcPosition ) ) {
        break;
      }
      vec3 ssCurrentPosition = ndcToSS( ndcPosition );
      vec2 sampleHeight      = textureLod( depthBuffer, ssCurrentPosition.xy, 0 ).xy;
      if ( ( !homogeneousReversedDepth && sampleHeight.x < ssCurrentPosition.z && sampleHeight.y >= ssCurrentPosition.z ) ||
        ( homogeneousReversedDepth && sampleHeight.x > ssCurrentPosition.z && sampleHeight.y <= ssCurrentPosition.z ) ) {
        // во избежание множества артефактов, делаем проверку на допустимую погрешность в зависимости от расстояния до камеры
        vec3  wsSamplePosition = ssToWSPosition( vec3( ssCurrentPosition.xy, sampleHeight.x ) );
        float bias             = 0.0004f;
        if ( !homogeneousReversedDepth ) {
          bias /= 1.0f - ssCurrentPosition.z;
        } else {
          bias /= ssCurrentPosition.z;
        }
        if ( length( wsSamplePosition - wsCurrentPosition ) < bias ) {
          continue;
        }
        return 0.0f;
        float shadowDepth = abs( sampleHeight.x - sampleHeight.y );
        float backBorder  = shadowDepth / abs( sampleHeight.x - ssCurrentPosition.z );
        float frontBorder = shadowDepth / abs( sampleHeight.y - ssCurrentPosition.z );
        // используем квадратичное затухание для уменьшения интенсивности тени с расстоянием
        // выборки, ближайшие к точке, имеют больший вес
        //float weigthCoeff = quadraticAttenuation( i / numSteps, 1.0f, lightAttenuationConstant );
        float weigthCoeff = 1.0f - ( i / numSteps );
        shadowMultiplier  = max( shadowMultiplier, /*min( frontBorder, backBorder ) */ weigthCoeff );
        ++numSamplesUnderSurface;
      }
    }
    if ( numSamplesUnderSurface != 0 ) {
      // используем самый темный из получившихся результатов
      shadowCoefficient = min( shadowCoefficient, clamp( 1.0f - shadowMultiplier, 0.0f, 1.0f ) );
    }
  }
  return shadowCoefficient;

  /*
Occlusion information generation
 Compute and store bent normal N' during SSAO pass
 Bent normal is average unoccluded direction
 Requires clean SSAO without any self-occlusion and a relatively wide radius
 For each light
 Compute N dot L as usual
 Compute N' dot L
 Center depth is full resolution, all other taps are FP16 half-resolution depth
 Attenuate lighting with the occlusion amount multiplied by a clamped
difference between the two dot products
*/

  /*
	// 77 ms
	float radius = 10.0f;
	int numSteps = 32;
	float sampleStep = radius / numSteps;

	vec3 wsSampleStepPosition;
	for ( int i = 1; i <= numSteps; ++i ) {
		wsSampleStepPosition = wsPosition + ( i * sampleStep * wsFragToLightDir );


		vec3 ssSampleStepPosition = wsToSSPosition( wsSampleStepPosition );

		if (textureLod( depthBuffer, ssSampleStepPosition.xy, 0 ).x < ssSampleStepPosition.z) {
            shadow = vec3(0.0f);
		}
	}
	return shadow;
*/

  /*
	// 50.3 ms
	vec3 ssPosition = getSSPosition( gl_FragCoord.xy, texelFetch( depthBuffer, ivec2( gl_FragCoord.xy ), 0 ).x );

	vec3 ssFragToLightDir = normalize( ssLightPosition - ssPosition );

	float radius = 0.125f;
	float numSteps = 32.0f;
	float sampleStep = radius / numSteps;
	vec3 positionStep = ssFragToLightDir * sampleStep;
	vec3 ssCurrentPosition = ssPosition;

	float shadowMultiplier = 0.0f;
	int numSamplesUnderSurface = 0;

	for ( int i = 0; i < numSteps; ++i ) {
		ssCurrentPosition += positionStep;
		//if (ssCurrentPosition.z >= 1) {
		//	return vec3(0,1,0.0f);
		//}
		//if ( length( ssCurrentPosition - ssPosition ) >= lightDistance ) {
		//	break;
		//}
		float sampleHeight = textureLod( depthBuffer, ssCurrentPosition.xy, 0 ).x;
		if (
			( !homogeneousReversedDepth && sampleHeight < ssCurrentPosition.z ) ||
			( homogeneousReversedDepth && sampleHeight > ssCurrentPosition.z )
		) {
			// TODO: сверять со вторым слоем
			float difference = abs( ssCurrentPosition.z - sampleHeight );
			if ( difference <= 0.001f ) {
            	return vec3(0.0f);
				++numSamplesUnderSurface;
				float strengthCoeff = 1-(i * sampleStep);

				float newShadowMultiplier = difference * strengthCoeff;
				shadowMultiplier = max( shadowMultiplier, newShadowMultiplier );
        	}
		}
	}
*/

  /*for ( int i = 1; i < numSteps; ++i ) {
		float stepCoeff = i * sampleStep;
		vec3 ssStepPosition = ssPosition + ( stepCoeff * ssFragToLightDir );
		float heightSample = textureLod( depthBuffer, ssStepPosition.xy, 0 ).x;


		// если сэмпл закрыт геометрией, считаем степень его затененности
		if (
			( homogeneousReversedDepth && heightSample > ssStepPosition.z ) ||
			( !homogeneousReversedDepth && heightSample < ssStepPosition.z )
		) {
			return vec3(0.0f);
			++numSamplesUnderSurface;
			float newShadowMultiplier = ( ssStepPosition.z - heightSample ) * ( 1.0f - stepCoeff );
			shadowMultiplier = max( shadowMultiplier, newShadowMultiplier );
		}
	}

	// Shadowing factor should be 1 if there were no points under the surface
	if ( numSamplesUnderSurface == 0 ) {
		return vec3(1.0f);
	}
	return vec3(1.0f - shadowMultiplier);*/
}

float calculateShadow(in vec3 wsPosition) {
  // clip space
  vec4 lsPosition = lightProjectionMatrix * vec4(wsPosition, 1.0f );
  // NDC
  vec3 projCoords = lsPosition.xyz / lsPosition.w;
  projCoords.xy = projCoords.xy * 0.5f + 0.5f;
  if ( !homogeneousReversedDepth ) {
    // если clip space в диапазоне [-1, 1], преобразуем в [0, 1]
    projCoords.z = projCoords.z * 0.5f + 0.5f;
  }
  return texture( shadowBuffer, projCoords );
}

vec3 calculateLight( in vec3 wsPosition, in vec3 wsNormal, in vec3 wsFragToLight, in float lightDistance ) {
  vec3 wsFragToLightDir = normalize( wsFragToLight );
  // рассеянный свет
  float diffuseCoeff = max( dot( wsNormal, wsFragToLightDir ), 0.0f );
  // учитываем искривленные нормали (хоть это и хак)
  if ( drawSSAO ) {
    vec3 wsBentNormal = texelFetch( bentNormalBuffer, ivec2( gl_FragCoord.xy ), 0 ).xyz;
    diffuseCoeff      = max( dot( wsBentNormal, wsFragToLightDir ), diffuseCoeff );
  }
  vec3 diffuse = diffuseCoeff * lightColor;

  if ( !drawLights ) {
  	return diffuse;
  }
  if ( lightType == 2 ) {
    // если spot
    float theta     = dot( wsFragToLightDir, normalize( -wsLightDirection ) );
    float epsilon   = lightConeInner - lightConeOuter;
    float intensity = clamp( ( theta - lightConeOuter ) / epsilon, 0.0f, 1.0f );
    diffuse *= intensity;
  } else if ( lightType == 3 ) {
    // если directional
    diffuse *= max( dot( wsFragToLightDir, normalize( -wsLightDirection ) ), 0.0f );
  }

  // по умолчанию тени нет
  float shadowCoefficient = 1.0f;
  if ( drawShadows ) {
    // тень
    if ( lightType == 1 ) {
      shadowCoefficient = calculateCubemapShadow( wsPosition, wsFragToLight, wsFragToLightDir, lightDistance );
    } else {
      shadowCoefficient = calculateShadow( wsPosition );
    }
    diffuse *= shadowCoefficient;
  }
  // считаем затухание только для участков, находящихся не в тени
  if ( lightType != 3 && shadowCoefficient != 0.0f ) {
    //diffuse *= quadraticAttenuationFast( lightDistance, lightRadius );
    diffuse *= quadraticAttenuation( lightDistance, lightRadius, lightAttenuationConstant );
  }
  return diffuse;
}

void main() {
  vec3  wsPosition    = getWSPositionFromDepth( gl_FragCoord.xy );
  vec3  wsFragToLight = wsLightPosition - wsPosition;
  float lightDistance = length( wsFragToLight );

  // считаем только те фрагменты, которые находятся на освещаемой источником поверхности
  if ( lightDistance > lightRadius ) {
    discard;
  }
  //vec3 wsNormal = decode( texelFetch( normalBuffer, ivec2( gl_FragCoord.xy ), 0 ).xy );
  vec3 wsNormal = texelFetch( normalBuffer, ivec2( gl_FragCoord.xy ), 0 ).xyz;

  /*vec3 ssFragToLight = ssLightPosition - ssPosition;
	if ( dot( ssNormal, ssFragToLight ) <= 0.0f ) {
		discard;
	}*/

  // и повернуты к нему лицом
  //if ( dot( wsNormal, wsFragToLight ) <= 0.0f ) {
//    discard;
  //}
  //fragmentColor = vec3(0,1,0);
  fragmentColor = calculateLight( wsPosition, wsNormal, wsFragToLight, lightDistance );
}
