#version 150 core
#extension GL_ARB_explicit_attrib_location : enable

uniform sampler2D normalBuffer;
uniform sampler2D depthBuffer;

uniform vec2 screenSize;
uniform vec3 wsCameraPosition;

uniform bool homogeneousReversedDepth;
uniform mat4 viewProjectionInversedMatrix;

uniform mat4 viewProjectionMatrix;

uniform int numSamples;

uniform float projectedRadius;

/** World-space AO radius in scene units (r).  e.g., 1.0m */
uniform float radius;
uniform float negInvRadius2;

/** Bias to avoid AO in smooth corners, e.g., 0.01m */
uniform float bias;
uniform float intensity;

// Ouput data
layout( location = 0 ) out float visibility;
layout( location = 1 ) out vec3 bentNormal;

// Returns +-1
vec2 signNotZero( in vec2 v ) {
  return vec2( ( v.x >= 0.0 ) ? 1.0 : -1.0, ( v.y >= 0.0 ) ? 1.0 : -1.0 );
}
// Assume normalized input. Output is on [-1, 1] for each component.
vec2 encode( in vec3 v ) {
  // Project the sphere onto the octahedron, and then onto the xy plane
  vec2 p = v.xy * ( 1.0 / ( abs( v.x ) + abs( v.y ) + abs( v.z ) ) );
  // Reflect the folds of the lower hemisphere over the diagonals
  return ( v.z <= 0.0 ) ? ( ( 1.0 - abs( p.yx ) ) * signNotZero( p ) ) : p;
}
vec3 decode( in vec2 e ) {
  vec3 v = vec3( e.xy, 1.0 - abs( e.x ) - abs( e.y ) );
  if ( v.z < 0 ) {
    v.xy = ( 1.0 - abs( v.yx ) ) * signNotZero( v.xy );
  }
  return normalize( v );
}

// получаем однородные координаты экрана из координат окна
vec2 getSSPosition( in vec2 windowSpacePosition ) {
  return windowSpacePosition / screenSize;
}
// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec4 getWSPosition( in vec2 ssPosition, in float z ) {
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    z = z * 2.0f - 1.0f;
  }
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssPosition * 2.0f - 1.0f, z, 1.0f );
  return wsPosition / wsPosition.w;
}

vec2 VogelDiskSample( int sampleIndex, float spinAngle, in float radiusOffset ) {
  const float goldenAngle = 2.39996f;
  float       theta       = sampleIndex * goldenAngle + spinAngle;
  return radiusOffset * vec2( cos( theta ), sin( theta ) );
}
/*vec2 tapLocation( int sampleIndex, float spinAngle, out float ssR ) {
	const float goldenAngle = 2.4f;

	float r = (float(sampleIndex) + 0.5) / numSamples;
	float theta = r * goldenAngle + spinAngle;

	ssR = r;

	return r * vec2(cos(theta), sin(theta));
}*/

/** Smaller return value = less occlusion */
float fallOffFunction( float vv, float vn ) {
  const float epsilon = 0.0002f;
  float       f       = clamp( vv * negInvRadius2 + 1.0f, 0.0f, 1.0f );
  return f * clamp( vn - bias, 0.0f, 1.0f ) / ( epsilon + vv );
}

/** Compute the occlusion due to sample point \a Q about camera-space point \a wsPosition with unit normal \a wsNormal */
float aoValueFromPositionsAndNormal( vec3 wsPosition, vec3 wsNormal, vec3 Q ) {
  vec3  v  = Q - wsPosition;
  float vv = dot( v, v );
  float vn = dot( v, wsNormal );

  float f = fallOffFunction( vv, vn );
  return f;
  // Enhance dark areas
  //return f * mix(0.9 + 0.5 * intensity, 1.2 - intensity * 0.15, f);
}

/** Read the camera-space position of the point at screen-space pixel ssP + unitOffset * ssR.  Assumes length(unitOffset) == 1. */
void getOffsetPositions( ivec2 ssCoord, vec2 unitOffset, float ssR, out vec3 P1, out vec3 P2 ) {
  ivec2 ssP = ivec2( ssR * unitOffset ) + ssCoord;

  // Offset to pixel center
  vec2 texCoords   = getSSPosition( vec2( ssP ) + 0.5f );
  vec2 depthValues = textureLod( depthBuffer, texCoords, 0 ).xy;
  P1               = getWSPosition( texCoords, depthValues.x ).xyz;
  P2               = getWSPosition( texCoords, depthValues.y ).xyz;
}

/** Compute the occlusion due to sample with index \a i about the pixel at \a ssCoord that corresponds
	to camera-space point \a wsPosition with unit normal \a wsNormal, using maximum screen-space sampling radius \a ssDiskRadius

	Note that units of H() in the HPG12 paper are meters, not
	unitless.  The whole falloff/sampling function is therefore
	unitless.  In this implementation, we factor out (9 / radius).

	Four versions of the falloff function are implemented below

	When sampling from the peeled depth buffer, make sure ssDiskRadius has been premultiplied by cszBufferScale
*/
float sampleAO( in ivec2 ssCoord, in vec3 wsPosition, in vec3 wsNormal, in float ssDiskRadius, in int sampleIndex,
  in float randomPatternRotationAngle, in mat3 tbnMatrix, out vec3 averageNormal, out vec3 Q1, out vec3 Q2 ) {
  //float diskRadiusOffset = float(sampleIndex + mod(randomPatternRotationAngle, 1.0) + 0.5) / (numSamples - 0.5);
  //vec2 unitOffset = tapLocation(diskRadiusOffset, randomPatternRotationAngle);
  float diskRadiusOffset = sqrt( ( sampleIndex + 0.5f ) / numSamples );
  vec2  unitOffset       = VogelDiskSample( sampleIndex, randomPatternRotationAngle, diskRadiusOffset );

  // 161.6
  // компромиссный вариант
  vec3 tsSamplePosition = diskRadiusOffset * vec3( unitOffset, clamp( 1.1f - diskRadiusOffset, 0.0f, 1.0f ) );
  vec3 wsSamplePosition = tbnMatrix * tsSamplePosition;

  wsSamplePosition = wsPosition + radius * wsSamplePosition;

  // project sample position (to sample texture) (to get position on screen/texture)
  vec4 offset = viewProjectionMatrix * vec4( wsSamplePosition, 1.0f );  // from world to clip-space
  offset.xy /= offset.w;                                                // perspective divide
  offset.xy = offset.xy * 0.5 + 0.5;                                    // transform to range 0.0 - 1.0

  //int mipLevel = clamp( int( log2( length( offset.xy * screenSize - ssCoord ) ) ) - logMaxOffset, 0, maxMipLevel );
  //vec3 wsSceneSample = getWSPosition( offset.xy, textureLod( depthBuffer, offset.xy, mipLevel ).r ).xyz;//169ms
  // если мип не считать и всегда брать из нулевого, получается быстрее всего. качество не страдает
  vec2 sceneDepth     = textureLod( depthBuffer, offset.xy, 0 ).xy;
  vec3 wsSceneSample1 = getWSPosition( offset.xy, sceneDepth.x ).xyz;
  vec3 wsSceneSample2 = getWSPosition( offset.xy, sceneDepth.y ).xyz;

  float visibilityCoeff = step( dot( wsSceneSample1 - wsSamplePosition, wsSceneSample1 - wsCameraPosition ), 0 );
  //float depthDifference = length( wsSceneSample - wsSamplePosition );
  //float rho = clamp( ( depthDifference - radius ) / depthDifference, 0.0, 1.0 );
  //float rangeCheck = smoothstep(0.9, 1.0, radius / length( wsSceneSample - wsPosition ));
  //float rangeCheck = length( wsSceneSample - wsPosition ) > radius ? 0.0f : 1.0f;

  vec3  v1         = wsSceneSample1 - wsPosition;
  float vv1        = dot( v1, v1 );
  float rangeCheck = clamp( vv1 * negInvRadius2 + 1.5f, 0.0f, 1.0f );  // 157

  // добавляем вклад второго слоя
  vec3  v2   = wsSceneSample2 - wsPosition;
  float vv2  = dot( v2, v2 );
  rangeCheck = max( rangeCheck, clamp( vv2 * negInvRadius2 + 1.5f, 0.0f, 1.0f ) );

  // изогнутые нормали
  float bentCoeff = rangeCheck * visibilityCoeff;
  averageNormal += normalize( wsSamplePosition - wsPosition ) * ( 1.0f - bentCoeff );

  float       vn          = dot( v1, wsNormal );
  const float epsilon     = 0.0002f;
  float       normalCheck = clamp( vn - bias, 0.0f, 1.0f ) / ( epsilon + vv1 );  // 162

  // Enhance dark areas
  float f = normalCheck * bentCoeff;
  return f * mix( 0.9 + 0.5 * intensity, 1.2 - intensity * 0.15, f );

  /*
	// это настоящий ssrt, но очень жрущий 340ms, результат выглядит не намного лучше классического ssao
	vec3 tsSamplePosition = diskRadiusOffset * vec3( unitOffset, 1.0 );
	vec3 wsSamplePosition = tbnMatrix * tsSamplePosition;

	int numSteps = 4;
	float radiusStep = radius / numSteps;
	float aoCoeffStep = 1.0f / numSteps;

	float aoCoeff = 0.0;

	vec3 wsSampleStepPosition;
	for ( int i = 1; i <= numSteps; ++i ) {
		wsSampleStepPosition = wsPosition + ( i * radiusStep * wsSamplePosition );

		// project sample position (to sample texture) (to get position on screen/texture)
		vec4 offset = viewProjectionMatrix * vec4( wsSampleStepPosition, 1.0f ); // from world to clip-space
		offset.xy /= offset.w; // perspective divide
		offset.xy = offset.xy * 0.5 + 0.5; // transform to range 0.0 - 1.0

		vec3 wsSceneSample = getWSPosition( offset.xy, texelFetch(depthBuffer, ivec2( offset.xy * screenSize ), 0 ).x ).xyz;

		// если сэмпл закрыт геометрией, считаем его невидимым и прекращаем обход
		if (dot( wsSceneSample - wsSampleStepPosition, wsSceneSample - wsCameraPosition ) < 0) {
			if ( length( wsSceneSample - wsPosition ) > radius ) {
				aoCoeff = 1.0f;
			}
			break;
		}
		aoCoeff += aoCoeffStep;

	}
	averageNormal += normalize( wsSampleStepPosition - wsPosition ) * aoCoeff;
	return 1.0 - aoCoeff;
*/
  /*
	// алгоритм основанный на ssao hemisphere, дает хорошие нормали
	vec3 tsSamplePosition = diskRadiusOffset * vec3( unitOffset, clamp( 1.0f - diskRadiusOffset + bias, 0.0f, 1.0f ) );
	//vec3 tsSamplePosition = diskRadiusOffset * vec3( 0.0, 0.0, 1.0f );
	vec3 wsSamplePosition = tbnMatrix * tsSamplePosition;

	wsSamplePosition = wsPosition + radius * wsSamplePosition;

	// project sample position (to sample texture) (to get position on screen/texture)
	vec4 offset = viewProjectionMatrix * vec4( wsSamplePosition, 1.0f ); // from world to clip-space
	offset.xy /= offset.w; // perspective divide
	offset.xy = offset.xy * 0.5 + 0.5; // transform to range 0.0 - 1.0

	//int mipLevel = clamp( int( log2( length( offset.xy * screenSize - ssCoord ) ) ) - logMaxOffset, 0, maxMipLevel );
	//vec3 wsSceneSample = getWSPosition( offset.xy, textureLod( depthBuffer, offset.xy, mipLevel ).r ).xyz;//178ms
	// если мип не считать и всегда брать из нулевого, получается быстрее всего. насколько проседает качество?
	vec3 wsSceneSample = getWSPosition( offset.xy, textureLod( depthBuffer, offset.xy, 0 ).r ).xyz;//163ms; 159
	//vec3 wsSceneSample = getWSPosition( offset.xy, texelFetch( depthBuffer, ivec2( offset.xy * screenSize ), 0 ).r ).xyz;//164ms

    // range check & accumulate
    //float rangeCheck = smoothstep(0.0, 1.0, 1 / length( wsSceneSample - wsPosition ));
    //occlusion += (sampleDepth >= sample.z + bias ? 1.0 : 0.0) * rangeCheck;

	//float bentCoeff = max(0.0, dot( Q1 - wsSamplePosition, Q1 - wsCameraPosition ) );
	//float bentCoeff = length( wsSamplePosition - wsCameraPosition ) <= length( Q1 - wsCameraPosition ) ? 1 : 0;

	float rangeCheck = 1.0f;
	//rangeCheck = smoothstep(0.0, 1.0, radius / length( wsSceneSample - wsPosition ));
	//rangeCheck = length( wsSceneSample - wsPosition ) < radius ? 1.0 : 0.0;
	rangeCheck = smoothstep(0.0, 1.0, radius / length( wsSceneSample - wsPosition ));

	//float depthDifference = length( wsSceneSample - wsSamplePosition );
	//float rangeCheck = clamp( ( depthDifference - radius ) / depthDifference, 0.0, 1.0 );

	// видимость сэмпла
	float coeff = 1.0f;
	//coeff = ceil( max( 0.0, dot( wsSceneSample - wsSamplePosition, wsSceneSample - wsCameraPosition ) ) );
	//coeff = length( wsSceneSample - wsCameraPosition ) <= length( wsSamplePosition - wsCameraPosition ) ? 1.0 : 0.0;
	coeff = dot( wsSceneSample - wsSamplePosition, wsSceneSample - wsCameraPosition ) < 0 ? 1.0 : 0.0;

	float totCoef = coeff * rangeCheck;

	averageNormal += normalize( wsSamplePosition - wsPosition ) * ( 1.0 - totCoef );

	return totCoef;
*/
  /*
	// 140,6 ms
	// оригинальный алгоритм, он не совсем честный физически, но очень быстрый и дает хороший результат AO,
	// изогнутые нормали считать не умеет :(
	// Ensure that the taps are at least 1 pixel away
	diskRadiusOffset = max(0.75f, diskRadiusOffset * ssDiskRadius);

	// The occluding point in camera space
	getOffsetPositions(ssCoord, unitOffset, diskRadiusOffset, Q1, Q2 );

	float aoCoeff = 0.0;
	//if ( drawSecondGeometryLayer ) {
	//	aoCoeff = max(aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q1), aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q2));
	//} else {
		aoCoeff = aoValueFromPositionsAndNormal(wsPosition, wsNormal, Q1);
	//}
	averageNormal += normalize( Q1 - wsPosition ) * (1.0f - aoCoeff);
	return aoCoeff;
*/
}

const float MIN_RADIUS = 4.0f;  // pixels

void main() {
  // Pixel being shaded
  ivec2 ssCoord = ivec2( gl_FragCoord.xy );

  // World space point being shaded
  float csDepth    = texelFetch( depthBuffer, ssCoord, 0 ).x;
  vec3  wsPosition = getWSPosition( getSSPosition( gl_FragCoord.xy ), csDepth ).xyz;

  // расстояние от камеры до фрагмента сцены
  float distanceToFragment = length( wsPosition - wsCameraPosition );
  // Choose the screen-space sample radius
  // proportional to the projected area of the sphere
  float ssDiskRadius = projectedRadius / distanceToFragment;
  //vec3 wsNormal = decode( texelFetch( normalBuffer, ssCoord, 0 ).xy );
  vec3 wsNormal = texelFetch( normalBuffer, ssCoord, 0 ).xyz;
  if ( ssDiskRadius <= MIN_RADIUS ) {
    // There is no way to compute AO at this radius
    visibility = 1.0;
    bentNormal = wsNormal;
    return;
  }

  vec3 randomVec = normalize( wsCameraPosition - wsPosition );
  // create tbnMatrix change-of-basis matrix: from tangent-space to world-space
  vec3 tangent   = normalize( randomVec - wsNormal * dot( randomVec, wsNormal ) );
  vec3 bitangent = cross( wsNormal, tangent );
  mat3 tbnMatrix = mat3( tangent, bitangent, wsNormal );

  // Hash of pixel position as pseudo-random value
  float randomPatternRotationAngle = ( float( ( ( ssCoord.x & 0xffff ) * ( ( ssCoord.y & 0xff ) + 1 ) ) ^
                                       ( ssCoord.x >> 16 ) + ( ( ( ssCoord.x + 127 ) & 0xffff ) * ( ( ssCoord.y + 128 ) & 0xff ) ) ) ) *
    2.39996;

  float sum = 0.0;
  vec3  averageNormal;
  for ( int i = 0; i < numSamples; ++i ) {
    vec3 Q1;
    vec3 Q2;
    sum += sampleAO( ssCoord, wsPosition, wsNormal, ssDiskRadius, i, randomPatternRotationAngle, tbnMatrix, averageNormal, Q1, Q2 );
  }

  averageNormal /= numSamples;
  bentNormal = normalize( averageNormal );

  visibility = pow( clamp( 1.0 - sqrt( sum * ( 1.0 / numSamples ) ), 0.0f, 1.0f ), intensity );
  //old//visibility = pow( 1.0f - ( 3.0f * sum / numSamples ), intensity);
  //visibility = pow( 1.0 - (sum / numSamples), intensity);
  //visibility = 1-(sum / numSamples);
  //visibility = sum / numSamples;
  // visibility = 1.0f;
}
